#include "RBM.h"
#include <cstdlib>
#include <float.h>

using namespace arma;

void RBM::Train(const arma::mat &data, const arma::mat &validation,
                UnitType visible_type, UnitType hidden_type,
                int num_hidden,
                int mini_batches,
                int iterations,
                double learning_rate,
                double momentum_multiplier,
                PenaltyType weight_decay_penalty_type,
                double weight_decay_multiplier,
                double sparsity_target,
                double sparsity_multiplier)
{
    m_visible_type = visible_type;
    m_hidden_type = hidden_type;

    int N = data.n_cols;
    int num_visible = data.n_rows;
    int total_batches = (int)ceil((float)N/mini_batches);

    m_weights.resize(num_hidden, num_visible);
    m_visible_biases = zeros(num_visible);
    m_hidden_biases = zeros(num_hidden); // works with sparsity

    // Keep track of the best parameters
    mat best_weights = m_weights;
    mat best_visible_biases = m_visible_biases;
    mat best_hidden_biases = m_hidden_biases;
    double best_rmse = DBL_MAX;

    //m_hidden_biases = ones(num_hidden)*-4;
    //m_visible_biases = ones(num_visible)*-4;

    m_weights.randn();
    m_weights *= 0.01;

    mat momentum_weights = zeros(num_hidden, num_visible);
    vec momentum_visible_biases = zeros(num_visible);
    vec momentum_hidden_biases = zeros(num_hidden);

    double final_learning_rate = learning_rate*0.1;
    double m = -(learning_rate - final_learning_rate)/iterations;
    double c = learning_rate;

    // Weight decay
    mat weight_penalty = m_weights*0;

    // Sparsity
    //mat qcurrent;//, qold = 0, qstdev = 0;
    double qcurrent;

    mat sparsity_weight_penalty = m_weights*0;
    mat sparsity_bias_penalty = m_hidden_biases*0;

    for(int i=0; i < iterations; i++) {
        int start = (i % total_batches)*mini_batches;
        int end = start + mini_batches - 1;

        if(end >= N) {
            end = N - 1;
        }

        mat batch = data.cols(start, end);

        Gradients g = CD1(batch, m_weights, m_visible_biases, m_hidden_biases);

        if(weight_decay_multiplier > 0.0) {
            if(weight_decay_penalty_type == L1) {
                for(unsigned int ii=0; ii < m_weights.n_rows; ii++) {
                    for(unsigned int jj=0; jj < m_weights.n_cols; jj++) {
                        // defined derivative at 0 as +1, instead of undefined
                        if(m_weights(ii, jj) >= 0) {
                            weight_penalty(ii, jj) = weight_decay_multiplier;
                        }
                        else {
                            weight_penalty(ii, jj) = -weight_decay_multiplier;
                        }
                    }
                }
            }
            else if(weight_decay_penalty_type == L2) {
                weight_penalty = m_weights*weight_decay_multiplier;
            }
        }

        // sparsity
        if(m_hidden_type == SIGMOID && sparsity_multiplier > 0.0) {
            int n = end - start + 1;

            mat z = m_weights*batch + repmat(m_hidden_biases, 1, n);
            mat h = Sigmoid(z);

            double s = accu(h) / h.n_elem;
            qcurrent = s;

            //mat s = mean(h, 0);
            //cout << s.n_rows << "x" << s.n_cols << endl;
            //qcurrent = s;

            double dE_ds = s - sparsity_target;
            //mat _dE_ds = s - sparsity_target;
            //mat dE_ds = repmat(_dE_ds.t(), 1, m_weights.n_cols);
            //cout << dE_ds.n_rows << "x" << dE_ds.n_cols << endl;

            //double ds_dh = 1.0/h.n_elem; // dont need this
            mat dh_dz = h % (1 - h);
            mat dz_dw = batch;

            sparsity_weight_penalty = dE_ds * dh_dz * dz_dw.t();
            sparsity_bias_penalty = dE_ds * sum(dh_dz, 1);

            //sparsity_weight_penalty = (dE_ds * ds_dh) % (dh_dz * dz_dw.t());
            //sparsity_bias_penalty = (_dE_ds.t() * ds_dh) % sum(dh_dz, 1);

            sparsity_weight_penalty *= sparsity_multiplier;
            sparsity_bias_penalty *= sparsity_multiplier;
        }

        momentum_weights = momentum_multiplier*momentum_weights + g.weights - weight_penalty - sparsity_weight_penalty;
        momentum_hidden_biases = momentum_multiplier*momentum_hidden_biases + g.hidden_biases- sparsity_bias_penalty;
        momentum_visible_biases = momentum_multiplier*momentum_visible_biases + g.visible_biases;

        double lr = m*i + c;

        m_weights += momentum_weights*lr ;
        m_hidden_biases += momentum_hidden_biases*lr;
        m_visible_biases += momentum_visible_biases*lr;

        double rmse_training = RMSE(data);
        double rmse_validation = DBL_MAX;

        if(validation.n_elem > 0) {
            rmse_validation = RMSE(validation);
        }

        if(rmse_validation < best_rmse) {
            best_rmse = rmse_validation;
            best_weights = m_weights;
            best_visible_biases = m_visible_biases;
            best_hidden_biases = m_hidden_biases;
        }

        if((i+1) % 10 == 0) {
            cout << "epoch: " << (i+1) << "/" << iterations;

            if(validation.n_elem > 0) {
                cout << " training/validation rmse: " << rmse_training << " " << rmse_validation << endl;
            }
            else {
                cout << " training rmse: " << rmse_training << endl;
            }

            if(weight_decay_multiplier > 0.0) {
                if(weight_decay_penalty_type == L1) {
                    double wd = accu(abs(m_weights));
                    cout << "    weight decay sum L1 penalty: " << wd << endl;
                }
                else if(weight_decay_penalty_type == L2) {
                    double wd = accu(m_weights % m_weights);
                    cout << "    weight decay sum L2 penalty: " << wd << endl;
                }
            }

            if(sparsity_multiplier > 0.0) {
                //mat avg = mean(qcurrent);
               cout << "    avg sparsity: " << qcurrent << endl;
            }
        }
    }

    if(validation.n_elem > 0) {
        m_weights = best_weights;
        m_visible_biases = best_visible_biases;
        m_hidden_biases = best_hidden_biases;
    }
}

arma::mat RBM::Round(const arma::mat &data)
{
    mat ret = data;

    for(unsigned int i=0; i < data.n_elem; i++) {
        ret(i) = round(data(i));
    }

    return ret;
}

arma::mat RBM::FeedForward(const arma::mat &data)
{
    int N = data.n_cols;

    m_input = data;
    m_output = m_weights*m_input + repmat(m_hidden_biases, 1, N);

    if(m_hidden_type == SIGMOID) {
        m_output = Sigmoid(m_output);
    }

    return m_output;
}

arma::mat RBM::Binarise(const arma::mat &input)
{
    mat output = zeros(input.n_rows, input.n_cols);

    for(unsigned int i=0; i < input.n_rows; i++) {
        for(unsigned int j=0; j < input.n_cols; j++) {
            if(input(i,j) >= 0.5) {
                output(i,j) = 1.0;
            }
            else {
                output(i,j) = 0.0;
            }
        }
    }

    return output;
}

arma::mat RBM::SampleBinary(const arma::mat &input)
{
    mat output = zeros(input.n_rows, input.n_cols);

    for(unsigned int i=0; i < input.n_rows; i++) {
        for(unsigned int j=0; j < input.n_cols; j++) {
            double r = (double)rand()/RAND_MAX;

            if(input(i,j) >= r) {
                output(i,j) = 1.0;
            }
            else {
                output(i,j) = 0.0;
            }
        }
    }

    return output;
}

arma::mat RBM::AddGaussianNoise(const arma::mat &M)
{
    mat ret = M + randn(M.n_rows, M.n_cols);

    return ret;
}

arma::mat RBM::Sigmoid(const arma::mat &M)
{
    mat ret(M.n_rows, M.n_cols);

    for(unsigned int i=0; i < M.n_rows; i++) {
        for(unsigned int j=0; j < M.n_cols; j++) {
            double x = M(i,j);
            ret(i,j) = 1.0/(1.0 + exp(-x));
        }
    }

    return ret;
}

arma::mat RBM::GradientWeights(const mat &visible_state, const mat &hidden_state)
{
    mat d_G_by_rbm_w = zeros(hidden_state.n_rows, visible_state.n_rows);
    int N = visible_state.n_cols;

    for(int i=0; i < N; i++) {
        d_G_by_rbm_w += hidden_state.col(i)*visible_state.col(i).t();
    }

    d_G_by_rbm_w /= N;

    return d_G_by_rbm_w;
}

arma::vec RBM::GradientVisibleBiases(const mat &visible_state, const vec &visible_bias)
{
    // For linear visible unit assume:
    // sigma = 1

    int N = visible_state.n_cols;

    vec d_G_by_vbias;

    if(m_visible_type == SIGMOID) {
        return mean(visible_state, 1);
    }
    else if(m_visible_type == LINEAR) {
        d_G_by_vbias = zeros(visible_bias.n_elem);

        for(int i=0; i < N; i++) {
            d_G_by_vbias += visible_state.col(i) - visible_bias; // deriv of -E, not E
        }
    }

    d_G_by_vbias /= N;

    return d_G_by_vbias;
}

arma::vec RBM::GradientHiddenBiases(const mat &hidden_state, const vec &hidden_bias)
{
    int N = hidden_state.n_cols;

    vec d_G_by_hbias;

    if(m_hidden_type == SIGMOID) {
        return mean(hidden_state, 1);
    }
    else if(m_hidden_type == LINEAR) {
        d_G_by_hbias = zeros(hidden_bias.n_elem);

        for(int i=0; i < N; i++) {
            d_G_by_hbias += hidden_state.col(i) - hidden_bias; // deriv of -E, not E
        }
    }

    d_G_by_hbias /= N;

    return d_G_by_hbias;
}

RBM::Gradients RBM::CD1(const mat &visible_data, const mat &weights, const vec &visible_biases, const vec &hidden_biases)
{
    // This versin of CD1 takes a few shortcuts from
    // "Using Very Deep Autoencoders for Content-Based Image Retrieval"

    mat visible_state, visible_probability;
    mat hidden_state, hidden_probability;
    mat gradient1, gradient2;
    vec visible_biases1, visible_biases2;
    vec hidden_biases1, hidden_biases2;

    Gradients ret;

    int N = visible_data.n_cols;

    // Positive phase
    visible_state = visible_data;

    // Can skip sampling, sometime it helps
    if(m_visible_type == SIGMOID) {
        visible_state = SampleBinary(visible_state);
    }
    else if(m_visible_type == LINEAR) {
        visible_state = AddGaussianNoise(visible_state);
    }

    if(m_hidden_type == SIGMOID) {
        hidden_probability = Sigmoid(weights*visible_state + repmat(hidden_biases, 1, N));
        hidden_state = SampleBinary(hidden_probability); // sample here, according to esann-deep-final.pdf
    }
    else if(m_hidden_type == LINEAR) {
        hidden_state = AddGaussianNoise(weights*visible_state + repmat(hidden_biases, 1, N));
    }

    gradient1 = GradientWeights(visible_state, hidden_state);
    visible_biases1 = GradientVisibleBiases(visible_state, visible_biases);
    hidden_biases1 = GradientHiddenBiases(hidden_state, hidden_biases);

    // Negative phase
    // Skip sampling here as well
    visible_state = weights.t()*hidden_state + repmat(visible_biases, 1, N);

    if(m_visible_type == SIGMOID) {
        visible_state = Sigmoid(visible_state);
        //visible_probability = Sigmoid(visible_state);
        //visible_state = SampleBinary(visible_probability);
    }
    else if(m_visible_type == LINEAR) {
        //visible_state = AddGaussianNoise(visible_state);
    }

    // skip sampling here
    if(m_hidden_type == SIGMOID) {
        hidden_probability = Sigmoid(weights*visible_state + repmat(hidden_biases, 1, N));
        hidden_state = hidden_probability;
    }
    else if(m_hidden_type == LINEAR) {
        hidden_state = weights*visible_state + repmat(hidden_biases, 1, N);
    }

    gradient2 = GradientWeights(visible_state, hidden_state);
    visible_biases2 = GradientVisibleBiases(visible_state, visible_biases);
    hidden_biases2 = GradientHiddenBiases(hidden_state, hidden_biases);

    // gradients
    ret.weights = gradient1 - gradient2;
    ret.visible_biases = visible_biases1 - visible_biases2;
    ret.hidden_biases= hidden_biases1 - hidden_biases2;

    return ret;
}

arma::mat RBM::Normalise(const arma::mat &data, double *ret_mean, double *ret_stdev)
{
    *ret_mean = accu(data) / data.n_elem;

    mat x = data - *ret_mean;
    mat x2 = x % x; // element-wise multiplication
    *ret_stdev = sqrt(accu(x2)/data.n_elem);

    return x/(*ret_stdev);
}

arma::mat RBM::dz_dh()
{
    return m_weights;
}

arma::mat RBM::dz_dw()
{
    return m_input;
}

arma::mat RBM::dh_dz()
{
    return (1.0 - m_output)%m_output;
}

arma::vec RBM::SampleVisible()
{
}

arma::vec RBM::SampleHidden()
{
    // pick some random visible state and run the markov chain
    vec visible_state = -1 + 2*randu(m_weights.n_cols);
    vec hidden_probability;
    vec hidden_state;

    for(int i=0; i < 1000; i++) {
        if(m_hidden_type == SIGMOID) {
            hidden_probability = Sigmoid(m_weights*visible_state + m_hidden_biases);
            hidden_state = hidden_probability; //SampleBinary(hidden_probability);
        }
        else if(m_hidden_type == LINEAR) {
            hidden_state = AddGaussianNoise(m_weights*visible_state + m_hidden_biases);
        }

        visible_state = m_weights.t()*hidden_state + m_visible_biases;

        if(m_visible_type == SIGMOID) {
            visible_state = Sigmoid(visible_state);
            visible_state = SampleBinary(visible_state);
        }
        else if(m_visible_type == LINEAR) {
            visible_state = AddGaussianNoise(visible_state);
        }
    }

    return hidden_state;
}

double RBM::RMSE(const arma::mat &data)
{
    // Reconstruction error, no sampling done, using raw probability
    int N = data.n_cols;

    mat hidden_state = Sigmoid(m_weights*data + repmat(m_hidden_biases, 1, N));
    mat reconstruction = m_weights.t()*hidden_state + repmat(m_visible_biases, 1, N);

    if(m_visible_type == SIGMOID) {
        reconstruction = Sigmoid(reconstruction);
    }

    mat err = data - reconstruction;
    err = err % err;
    double rmse = sqrt(accu(err)/err.n_elem);

    return rmse;
}
