#ifndef __RBM_H__
#define __RBM_H__

#include <armadillo>

// Remember to normalise your data before using this class, call RBM::Normalise()

class RBM
{
public:
    enum UnitType {SIGMOID, LINEAR};
    enum PenaltyType {L1, L2};

    void Train(const arma::mat &data, const arma::mat &validation,
               UnitType visible_type, UnitType hidden_type,
               int num_hidden,
               int mini_batches = 100,
               int iterations = 1000,
               double learning_rate = 0.01,
               double momentum_multiplier = 0.9,
               PenaltyType weight_decay_penalty_type = L1,
               double weight_decay_multiplier = 0.0,
               double sparsity_target = 0.01,
               double sparsity_multiplier = 0.0);

    arma::mat FeedForward(const arma::mat &data);
    arma::vec SampleVisible();
    arma::vec SampleHidden();

    static arma::mat Sigmoid(const arma::mat &M);
    static arma::mat Normalise(const arma::mat &data, double *ret_mean, double *ret_stdev);
    static arma::mat Round(const arma::mat &data);

    // partial derivatives
    arma::mat dz_dh();
    arma::mat dz_dw();
    arma::mat dh_dz(); // hidden deriv

public:
    UnitType m_visible_type;
    UnitType m_hidden_type;

    // Accessible for convenience
    arma::mat m_weights;
    arma::mat m_visible_biases;
    arma::mat m_hidden_biases;

    // Keep copies of useful data for backprop when doing forward pass
    arma::mat m_input;
    arma::mat m_output;

private:
    struct Gradients
    {
        arma::mat weights;
        arma::mat visible_biases;
        arma::mat hidden_biases;
    };

    static arma::mat Binarise(const arma::mat &input);
    static arma::mat SampleBinary(const arma::mat &input);
    static arma::mat AddGaussianNoise(const arma::mat &M); // matrix values are the mean

    arma::mat GradientWeights(const arma::mat &visible_state, const arma::mat &hidden_state);
    arma::vec GradientVisibleBiases(const arma::mat &visible_state, const arma::vec &visible_bias);
    arma::vec GradientHiddenBiases(const arma::mat &hidden_state, const arma::vec &hidden_bias);
    Gradients CD1(const arma::mat &visible_data, const arma::mat &weights, const arma::vec &visible_bias, const arma::vec &hidden_bias);

    double RMSE(const arma::mat &data);

private:
};

#endif
