#include <iostream>
#include <armadillo>
#include <cmath>
#include <cassert>
#include <vector>
#include <sstream>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "RBM.h"
#include "AutoEncoder.h"

using namespace std;
using namespace arma;

mat LoadData(const char *filename);
void DisplayWeights(const mat &weights);
void DisplayReconstruction(const mat &data, const mat &reconstruction);
void SaveWeightImages(const char *filename, const mat &weights);
void SaveReconstruction(const char *filename, const mat &data, const mat &reconstruction);

// Dataset
const char DATASET_FILE[] = "/home/nghia/Downloads/cifar-10-batches-bin/data_batch_1.bin";
const unsigned int MAX_SAMPLES = 5000; // limit the number of samples to train
const int DOWNSAMPLE_SIZE = 16; // original is 32

int main()
{
    // RBM
    // These two variables must be the same length
    const int rbm_layers[] = {-1, 100}; // first value is ignored, it is the input data
    const RBM::UnitType rbm_layer_types[] = {RBM::LINEAR, RBM::SIGMOID};

    const int num_rbms = sizeof(rbm_layers) / sizeof(int) - 1;
    const int mini_batches = 100;
    const int iterations = 1000;
    const double learning_rate = 0.05;
    const double momentum_multiplier = 0.9;

    // Weight decay penalty
    const RBM::PenaltyType weight_penalty_type = RBM::L1;
    const double weight_decay_multiplier = 0.0; // 0 to turn it off

    // Sparsity penalty
    const double sparsity_target = 0.01;
    const double sparsity_multiplier = 1.0; // 0 to turn it off

    assert(sizeof(rbm_layers) == sizeof(rbm_layer_types));

    vector <RBM> rbms(num_rbms);
    cv::namedWindow("main");

    mat data = LoadData(DATASET_FILE);
    mat data_orig = data;

    // Normalise the data
    double u=0, s=0; // mean, stdev
    data = RBM::Normalise(data, &u, &s);
    mat data_train = data;

    cout << "Total training data: " << data.n_cols << endl;

    // Stack the RBMs and train them
    for(int i=0; i < num_rbms; i++) {
        rbms[i].Train(data_train, mat(),
                      rbm_layer_types[i], rbm_layer_types[i+1],
                      rbm_layers[i+1],
                      mini_batches,
                      iterations,
                      learning_rate,
                      momentum_multiplier,
                      weight_penalty_type, weight_decay_multiplier,
                      sparsity_target, sparsity_multiplier
                      );

        data_train = rbms[i].FeedForward(data_train);
    }

    // Refine the weights using an autoencoder network
    AutoEncoder autoencoder;

    autoencoder.SetRBMHalf(rbms); // this will mirror the RBMS
    autoencoder.BackPropagation(data, 500, 0.01, momentum_multiplier, true);

    // Reconstruction error
    mat reconstruction = autoencoder.FeedForward(data);
    reconstruction = reconstruction*s + u;

    mat err = data_orig - reconstruction;
    err = err % err; // err .* err
    double rmse = sqrt(accu(err) / err.n_elem);
    cout << "reconstruction error (RMSE): " << rmse << endl;

    //DisplayWeights(autoencoder.m_rbms[0].m_weights);
    //DisplayReconstruction(data_orig, reconstruction);
    SaveWeightImages("weights.png", autoencoder.m_rbms[0].m_weights);
    SaveReconstruction("reconstruction.png", data_orig, reconstruction);

    return 0;
}

mat LoadData(const char *filename)
{
    const int image_size = 32;

    mat data;
    char label;
    vector <unsigned char> buffer(image_size*image_size*3); // temp buffer for RGB patch

    // Load the data as column vectors
    ifstream input(filename);

    if(!input) {
        cerr << "Error opening dataset: " << filename << endl;
        exit(1);
    }

    while(input.read((char*)&label, sizeof(label))) {
        input.read((char*)&buffer[0], buffer.size());

        if(data.n_cols < MAX_SAMPLES) {
            cv::Mat img(image_size, image_size, CV_8UC3);

            for(int y=0; y < image_size; y++) {
                for(int x=0; x < image_size; x++) {
                    img.at<cv::Vec3b>(y,x)[2] = buffer[y*image_size+x];
                    img.at<cv::Vec3b>(y,x)[1] = buffer[image_size*image_size + y*image_size+x];
                    img.at<cv::Vec3b>(y,x)[0] = buffer[image_size*image_size*2 + y*image_size+x];
                }
            }

/*
            cv::resize(img, img, cv::Size(320,320), 0, 0, cv::INTER_NEAREST);
            cv::imshow("main", img);
            cv::waitKey(0);
*/

            cv::cvtColor(img, img, CV_BGR2GRAY);
            cv::resize(img, img, cv::Size(DOWNSAMPLE_SIZE, DOWNSAMPLE_SIZE), 0, 0, cv::INTER_LINEAR);

            vec col_data(DOWNSAMPLE_SIZE*DOWNSAMPLE_SIZE);

            for(int i=0; i < DOWNSAMPLE_SIZE*DOWNSAMPLE_SIZE; i++) {
                col_data(i) = img.at<uchar>(i);
            }

            data = join_rows(data, col_data);


/*
            // Grab 8x8 patches
            for(int i=0; i < 4; i++) {
                for(int j=0; j < 4; j++) {
                    cv::Mat patch = img(cv::Rect(i*8, j*8, 8, 8));
                    vec col_data(8*8);

                    for(int i=0; i < 8*8; i++) {
                        col_data(i) = patch.at<uchar>(i);
                    }

                    data = join_rows(data, col_data);

                    //cv::resize(patch, patch, cv::Size(320,320), 0, 0, cv::INTER_NEAREST);
                    //cv::imshow("main", patch);
                    //cv::waitKey(0);

                }
            }
*/
        }
    }

    return data;
}

void DisplayWeights(const mat &weights)
{
    // Scale so that negative values range from [0, 128] and positive from [128, 255]

    int width = sqrt(weights.n_cols); // assume square image

    mat a = abs(weights);
    double max_val = a.max();
    cout << "min: " << weights.min() << endl;
    cout << "max: " << weights.max() << endl;

    cout << "weights.n_cols: " << weights.n_cols << endl;

    for(unsigned int j=0; j < weights.n_rows; j++) {
        mat tmp = (weights.row(j) / max_val)*127 + 128;

        cv::Mat img(width, width, CV_8U);

        for(int k=0; k < width*width; k++) {
            int v = tmp(k);

            if(v < 0) v = 0;
            if(v >= 256) v = 255;

            img.at<uchar>(k) = v;
        }

        cv::resize(img, img, cv::Size(width*20, width*20), 0, 0, cv::INTER_NEAREST);
        cv::imshow("main", img);
        cv::waitKey(0);
    }
}

void SaveWeightImages(const char *filename, const mat &weights)
{
    int image_width = sqrt(weights.n_cols);
    int images_per_row = sqrt(weights.n_rows) + 0.5;
    int scale = 4;
    int gap = 2;

    int canvas_width = (image_width+gap)*images_per_row;
    int canvas_height = canvas_width;

    cv::Mat canvas = cv::Mat::zeros(canvas_height, canvas_width, CV_8U);

    mat tmp = abs(weights);
    double max_val = tmp.max();

    int k=0;
    for(int y=0; y < images_per_row; y++) {
        for(int x=0; x < images_per_row; x++) {
            if(k >= (int)weights.n_rows) {
                y = images_per_row;
                break;
            }

            mat tmp = (weights.row(k) / max_val)*127 + 128;

            cv::Mat img(image_width, image_width, CV_8U);

            for(int i=0; i < image_width*image_width; i++) {
                int v = tmp(i);

                if(v < 0) v = 0;
                if(v >= 256) v = 255;

                img.at<uchar>(i) = v;
            }

            img.copyTo(canvas(cv::Rect(x*(image_width+gap), y*(image_width+gap), image_width, image_width)));

            k++;
        }
    }

    cv::resize(canvas, canvas, cv::Size(canvas_width*scale, canvas_width*scale), 0, 0, cv::INTER_NEAREST);
    cv::imwrite(filename, canvas);
}

void DisplayReconstruction(const mat &data, const mat &reconstruction)
{
    int image_size = DOWNSAMPLE_SIZE;
    int scale = 4;

    // Display the resuts
    for(unsigned int i=0; i < reconstruction.n_cols; i+=10) {
        cv::Mat canvas = cv::Mat::zeros(image_size, image_size*2+2, CV_8U);
        cv::Mat img1(image_size, image_size, CV_8U);
        cv::Mat img2(image_size, image_size, CV_8U);

        for(int j=0; j < image_size*image_size; j++) {
            img1.at<uchar>(j) = data(j,i);
        }

        for(int j=0; j < image_size*image_size; j++) {
            int v = reconstruction(j,i);

            if(v < 0) v = 0;
            if(v >= 256) v = 255;

            img2.at<uchar>(j) = v;
        }

        img1.copyTo(canvas(cv::Rect(0,0,image_size,image_size)));
        img2.copyTo(canvas(cv::Rect(image_size+1,0,image_size,image_size)));

        cv::resize(canvas, canvas, cv::Size((image_size*2+2)*scale,image_size*scale), 0, 0, cv::INTER_NEAREST);
        cv::imshow("main", canvas);
        cv::waitKey(0);
    }
}

void SaveReconstruction(const char *filename, const mat &data, const mat &reconstruction)
{
    int num_images = 10;
    int image_size = DOWNSAMPLE_SIZE;
    int gap = 2;
    int scale = 4;

    cv::Mat canvas = cv::Mat::zeros((image_size+gap)*num_images, (image_size+gap)*2, CV_8U);

    // Display the resuts
    for(unsigned int i=0; i < num_images; i++) {
        cv::Mat img1(image_size, image_size, CV_8U);
        cv::Mat img2(image_size, image_size, CV_8U);

        for(int j=0; j < image_size*image_size; j++) {
            img1.at<uchar>(j) = data(j,i);
        }

        for(int j=0; j < image_size*image_size; j++) {
            int v = reconstruction(j,i);

            if(v < 0) v = 0;
            if(v >= 256) v = 255;

            img2.at<uchar>(j) = v;
        }

        img1.copyTo(canvas(cv::Rect(0, i*(image_size+gap), image_size, image_size)));
        img2.copyTo(canvas(cv::Rect(image_size+gap, i*(image_size+gap), image_size, image_size)));
    }

    cv::resize(canvas, canvas, cv::Size(canvas.cols*scale, canvas.rows*scale), 0, 0, cv::INTER_NEAREST);
    cv::imwrite(filename, canvas);
}
